/*
 * World.cpp
 *
 *  Created on: Feb 3, 2010
 *      Author: crueger
 */

#include "Helpers/MemDebug.hpp"

#include "World.hpp"
#include "Patterns/ObservedContainer_impl.hpp"

#include "atom.hpp"
#include "config.hpp"
#include "molecule.hpp"
#include "periodentafel.hpp"
#include "ThermoStatContainer.hpp"
#include "Descriptors/AtomDescriptor.hpp"
#include "Descriptors/AtomDescriptor_impl.hpp"
#include "Descriptors/MoleculeDescriptor.hpp"
#include "Descriptors/MoleculeDescriptor_impl.hpp"
#include "Descriptors/SelectiveIterator_impl.hpp"
#include "Actions/ManipulateAtomsProcess.hpp"
#include "Helpers/Assert.hpp"
#include "Box.hpp"
#include "Matrix.hpp"
#include "defs.hpp"

#include "Patterns/Singleton_impl.hpp"

using namespace std;

/******************************* getter and setter ************************/
periodentafel *&World::getPeriode(){
  return periode;
}

config *&World::getConfig(){
  return configuration;
}

// Atoms

atom* World::getAtom(AtomDescriptor descriptor){
  return descriptor.find();
}

vector<atom*> World::getAllAtoms(AtomDescriptor descriptor){
  return descriptor.findAll();
}

vector<atom*> World::getAllAtoms(){
  return getAllAtoms(AllAtoms());
}

int World::numAtoms(){
  return atoms.size();
}

// Molecules

molecule *World::getMolecule(MoleculeDescriptor descriptor){
  return descriptor.find();
}

std::vector<molecule*> World::getAllMolecules(MoleculeDescriptor descriptor){
  return descriptor.findAll();
}

std::vector<molecule*> World::getAllMolecules(){
  return getAllMolecules(AllMolecules());
}

int World::numMolecules(){
  return molecules_deprecated->ListOfMolecules.size();
}

// system

Box& World::getDomain() {
  return *cell_size;
}

void World::setDomain(const Matrix &mat){
  *cell_size = mat;
}

void World::setDomain(double * matrix)
{
  Matrix M = ReturnFullMatrixforSymmetric(matrix);
  cell_size->setM(M);
}

std::string World::getDefaultName() {
  return defaultName;
}

void World::setDefaultName(std::string name)
{
  defaultName = name;
};

class ThermoStatContainer * World::getThermostats()
{
  return Thermostats;
}


int World::getExitFlag() {
  return ExitFlag;
}

void World::setExitFlag(int flag) {
  if (ExitFlag < flag)
    ExitFlag = flag;
}

/******************** Methods to change World state *********************/

molecule* World::createMolecule(){
  OBSERVE;
  molecule *mol = NULL;
  mol = NewMolecule();
  moleculeId_t id = getNextMoleculeId();
  ASSERT(!molecules.count(id),"proposed id did not specify an unused ID");
  mol->setId(id);
  // store the molecule by ID
  molecules[mol->getId()] = mol;
  mol->signOn(this);
  return mol;
}

void World::destroyMolecule(molecule* mol){
  OBSERVE;
  destroyMolecule(mol->getId());
}

void World::destroyMolecule(moleculeId_t id){
  OBSERVE;
  molecule *mol = molecules[id];
  ASSERT(mol,"Molecule id that was meant to be destroyed did not exist");
  DeleteMolecule(mol);
  molecules.erase(id);
  releaseMoleculeId(id);
}

atom *World::createAtom(){
  OBSERVE;
  atomId_t id = getNextAtomId();
  ASSERT(!atoms.count(id),"proposed id did not specify an unused ID");
  atom *res = NewAtom(id);
  res->setWorld(this);
  // store the atom by ID
  atoms[res->getId()] = res;
  return res;
}


int World::registerAtom(atom *atom){
  OBSERVE;
  atomId_t id = getNextAtomId();
  atom->setId(id);
  atom->setWorld(this);
  atoms[atom->getId()] = atom;
  return atom->getId();
}

void World::destroyAtom(atom* atom){
  OBSERVE;
  int id = atom->getId();
  destroyAtom(id);
}

void World::destroyAtom(atomId_t id) {
  OBSERVE;
  atom *atom = atoms[id];
  ASSERT(atom,"Atom ID that was meant to be destroyed did not exist");
  DeleteAtom(atom);
  atoms.erase(id);
  releaseAtomId(id);
}

bool World::changeAtomId(atomId_t oldId, atomId_t newId, atom* target){
  OBSERVE;
  // in case this call did not originate from inside the atom, we redirect it,
  // to also let it know that it has changed
  if(!target){
    target = atoms[oldId];
    ASSERT(target,"Atom with that ID not found");
    return target->changeId(newId);
  }
  else{
    if(reserveAtomId(newId)){
      atoms.erase(oldId);
      atoms.insert(pair<atomId_t,atom*>(newId,target));
      return true;
    }
    else{
      return false;
    }
  }
}

ManipulateAtomsProcess* World::manipulateAtoms(boost::function<void(atom*)> op,std::string name,AtomDescriptor descr){
  return new ManipulateAtomsProcess(op, descr,name,true);
}

ManipulateAtomsProcess* World::manipulateAtoms(boost::function<void(atom*)> op,std::string name){
  return manipulateAtoms(op,name,AllAtoms());
}

/********************* Internal Change methods for double Callback and Observer mechanism ********/

void World::doManipulate(ManipulateAtomsProcess *proc){
  proc->signOn(this);
  {
    OBSERVE;
    proc->doManipulate(this);
  }
  proc->signOff(this);
}
/******************************* IDManagement *****************************/

// Atoms

atomId_t World::getNextAtomId(){
  // try to find an Id in the pool;
  if(!atomIdPool.empty()){
    atomIdPool_t::iterator iter=atomIdPool.begin();
    atomId_t id = iter->first;
    pair<atomId_t,atomId_t> newRange = make_pair(id+1,iter->second);
    // we wont use this iterator anymore, so we don't care about invalidating
    atomIdPool.erase(iter);
    if(newRange.first<newRange.second){
      atomIdPool.insert(newRange);
    }
    return id;
  }
  // Nothing in the pool... we are out of luck
  return currAtomId++;
}

void World::releaseAtomId(atomId_t id){
  atomIdPool.insert(make_pair(id,id+1));
  defragAtomIdPool();
}

bool World::reserveAtomId(atomId_t id){
  if(id>=currAtomId ){
    pair<atomId_t,atomId_t> newRange = make_pair(currAtomId,id);
    if(newRange.first<newRange.second){
      atomIdPool.insert(newRange);
    }
    currAtomId=id+1;
    defragAtomIdPool();
    return true;
  }
  // look for a range that matches the request
  for(atomIdPool_t::iterator iter=atomIdPool.begin();iter!=atomIdPool.end();++iter){
    if(iter->first>id){
      // we have coverd all available ranges... nothing to be found here
      break;
    }
    // no need to check first, since it has to be <=id, since otherwise we would have broken out
    if(iter->second > id){
      // we found a matching range... get the id from this range

      // split up this range at the point of id
      pair<atomId_t,atomId_t> bottomRange = make_pair(iter->first,id);
      pair<atomId_t,atomId_t> topRange = make_pair(id+1,iter->second);
      // remove this range
      atomIdPool.erase(iter);
      if(bottomRange.first<bottomRange.second){
        atomIdPool.insert(bottomRange);
      }
      if(topRange.first<topRange.second){
        atomIdPool.insert(topRange);
      }
      defragAtomIdPool();
      return true;
    }
  }
  // this ID could not be reserved
  return false;
}

void World::defragAtomIdPool(){
  // check if the situation is bad enough to make defragging neccessary
  if((numAtomDefragSkips<MAX_FRAGMENTATION_SKIPS) &&
     (atomIdPool.size()<lastAtomPoolSize+MAX_POOL_FRAGMENTATION)){
    ++numAtomDefragSkips;
    return;
  }
  for(atomIdPool_t::iterator iter = atomIdPool.begin();iter!=atomIdPool.end();){
    // see if this range is adjacent to the next one
    atomIdPool_t::iterator next = iter;
    next++;
    if(next!=atomIdPool.end() && (next->first==iter->second)){
      // merge the two ranges
      pair<atomId_t,atomId_t> newRange = make_pair(iter->first,next->second);
      atomIdPool.erase(iter);
      atomIdPool.erase(next);
      pair<atomIdPool_t::iterator,bool> res = atomIdPool.insert(newRange);
      ASSERT(res.second,"Id-Pool was confused");
      iter=res.first;
      continue;
    }
    ++iter;
  }
  if(!atomIdPool.empty()){
    // check if the last range is at the border
    atomIdPool_t::iterator iter = atomIdPool.end();
    iter--;
    if(iter->second==currAtomId){
      currAtomId=iter->first;
      atomIdPool.erase(iter);
    }
  }
  lastAtomPoolSize=atomIdPool.size();
  numAtomDefragSkips=0;
}

// Molecules

moleculeId_t World::getNextMoleculeId(){
  // try to find an Id in the pool;
  if(!moleculeIdPool.empty()){
    moleculeIdPool_t::iterator iter=moleculeIdPool.begin();
    moleculeId_t id = iter->first;
    pair<moleculeId_t,moleculeId_t> newRange = make_pair(id+1,iter->second);
    // we wont use this iterator anymore, so we don't care about invalidating
    moleculeIdPool.erase(iter);
    if(newRange.first<newRange.second){
      moleculeIdPool.insert(newRange);
    }
    return id;
  }
  // Nothing in the pool... we are out of luck
  return currMoleculeId++;
}

void World::releaseMoleculeId(moleculeId_t id){
  moleculeIdPool.insert(make_pair(id,id+1));
  defragMoleculeIdPool();
}

bool World::reserveMoleculeId(moleculeId_t id){
  if(id>=currMoleculeId ){
    pair<moleculeId_t,moleculeId_t> newRange = make_pair(currMoleculeId,id);
    if(newRange.first<newRange.second){
      moleculeIdPool.insert(newRange);
    }
    currMoleculeId=id+1;
    defragMoleculeIdPool();
    return true;
  }
  // look for a range that matches the request
  for(moleculeIdPool_t::iterator iter=moleculeIdPool.begin();iter!=moleculeIdPool.end();++iter){
    if(iter->first>id){
      // we have coverd all available ranges... nothing to be found here
      break;
    }
    // no need to check first, since it has to be <=id, since otherwise we would have broken out
    if(iter->second > id){
      // we found a matching range... get the id from this range

      // split up this range at the point of id
      pair<moleculeId_t,moleculeId_t> bottomRange = make_pair(iter->first,id);
      pair<moleculeId_t,moleculeId_t> topRange = make_pair(id+1,iter->second);
      // remove this range
      moleculeIdPool.erase(iter);
      if(bottomRange.first<bottomRange.second){
        moleculeIdPool.insert(bottomRange);
      }
      if(topRange.first<topRange.second){
        moleculeIdPool.insert(topRange);
      }
      defragMoleculeIdPool();
      return true;
    }
  }
  // this ID could not be reserved
  return false;
}

void World::defragMoleculeIdPool(){
  // check if the situation is bad enough to make defragging neccessary
  if((numMoleculeDefragSkips<MAX_FRAGMENTATION_SKIPS) &&
     (moleculeIdPool.size()<lastMoleculePoolSize+MAX_POOL_FRAGMENTATION)){
    ++numMoleculeDefragSkips;
    return;
  }
  for(moleculeIdPool_t::iterator iter = moleculeIdPool.begin();iter!=moleculeIdPool.end();){
    // see if this range is adjacent to the next one
    moleculeIdPool_t::iterator next = iter;
    next++;
    if(next!=moleculeIdPool.end() && (next->first==iter->second)){
      // merge the two ranges
      pair<moleculeId_t,moleculeId_t> newRange = make_pair(iter->first,next->second);
      moleculeIdPool.erase(iter);
      moleculeIdPool.erase(next);
      pair<moleculeIdPool_t::iterator,bool> res = moleculeIdPool.insert(newRange);
      ASSERT(res.second,"Id-Pool was confused");
      iter=res.first;
      continue;
    }
    ++iter;
  }
  if(!moleculeIdPool.empty()){
    // check if the last range is at the border
    moleculeIdPool_t::iterator iter = moleculeIdPool.end();
    iter--;
    if(iter->second==currMoleculeId){
      currMoleculeId=iter->first;
      moleculeIdPool.erase(iter);
    }
  }
  lastMoleculePoolSize=moleculeIdPool.size();
  numMoleculeDefragSkips=0;
}

/******************************* Iterators ********************************/

// external parts with observers

CONSTRUCT_SELECTIVE_ITERATOR(atom*,World::AtomSet,AtomDescriptor);

World::AtomIterator
World::getAtomIter(AtomDescriptor descr){
    return AtomIterator(descr,atoms);
}

World::AtomIterator
World::getAtomIter(){
    return AtomIterator(AllAtoms(),atoms);
}

World::AtomIterator
World::atomEnd(){
  return AtomIterator(AllAtoms(),atoms,atoms.end());
}

// Internal parts, without observers

// Build the AtomIterator from template
CONSTRUCT_SELECTIVE_ITERATOR(atom*,World::AtomSet::set_t,AtomDescriptor);


World::internal_AtomIterator
World::getAtomIter_internal(AtomDescriptor descr){
  return internal_AtomIterator(descr,atoms.getContent());
}

World::internal_AtomIterator
World::atomEnd_internal(){
  return internal_AtomIterator(AllAtoms(),atoms.getContent(),atoms.end_internal());
}

// build the MoleculeIterator from template
CONSTRUCT_SELECTIVE_ITERATOR(molecule*,World::MoleculeSet,MoleculeDescriptor);

World::MoleculeIterator World::getMoleculeIter(MoleculeDescriptor descr){
  return MoleculeIterator(descr,molecules);
}

World::MoleculeIterator World::moleculeEnd(){
  return MoleculeIterator(AllMolecules(),molecules,molecules.end());
}

/******************************* Singleton Stuff **************************/

World::World() :
    Observable("World"),
    periode(new periodentafel),
    configuration(new config),
    Thermostats(new ThermoStatContainer),
    ExitFlag(0),
    atoms(this),
    currAtomId(0),
    lastAtomPoolSize(0),
    numAtomDefragSkips(0),
    molecules(),
    currMoleculeId(0),
    molecules_deprecated(new MoleculeListClass(this))
{
  cell_size = new Box;
  Matrix domain;
  domain.at(0,0) = 20;
  domain.at(1,1) = 20;
  domain.at(2,2) = 20;
  cell_size->setM(domain);
  defaultName = "none";
  molecules_deprecated->signOn(this);
}

World::~World()
{
  molecules_deprecated->signOff(this);
  delete cell_size;
  delete molecules_deprecated;
  delete periode;
  delete configuration;
  delete Thermostats;
  MoleculeSet::iterator molIter;
  for(molIter=molecules.begin();molIter!=molecules.end();++molIter){
    DeleteMolecule((*molIter).second);
  }
  molecules.clear();
  AtomSet::iterator atIter;
  for(atIter=atoms.begin();atIter!=atoms.end();++atIter){
    DeleteAtom((*atIter).second);
  }
  atoms.clear();
}

// Explicit instantiation of the singleton mechanism at this point

CONSTRUCT_SINGLETON(World)

/******************************* deprecated Legacy Stuff ***********************/

MoleculeListClass *&World::getMolecules() {
  return molecules_deprecated;
}
